/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.events;

import java.util.EventObject;

import de.bsvrz.sys.funclib.bitctrl.modell.MesswertDatum;
import de.bsvrz.sys.funclib.bitctrl.modell.OnlineDatensatz;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;

/**
 * Beschreibt ein Event, welches beim Eintreten neuer Analysedaten f&uuml;r ein
 * Systemobjekt ausgel&ouml;st wird.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class MesswertEvent extends EventObject {

	/** Die Eigenschaft {@code serialVersionUID}. */
	private static final long serialVersionUID = 1;

	/** Die betroffene Systemobjekt. */
	private final SystemObjekt objekt;

	/** Die aktuallisierten Messwerte. */
	private final OnlineDatensatz<? extends MesswertDatum> datensatz;

	/** Das aktuelle Datum. */
	private final MesswertDatum datum;

	/**
	 * Erzeugt ein neues Fuzzy-Event.
	 *
	 * @param source
	 *            Quelle des Events
	 * @param objekt
	 *            das betroffene Systemobjekt.
	 * @param datensatz
	 *            die aktuallisierten Messwerte des Objekts
	 * @param datum
	 *            das aktuelle Datum.
	 */
	public MesswertEvent(final Object source, final SystemObjekt objekt,
			final OnlineDatensatz<? extends MesswertDatum> datensatz,
			final MesswertDatum datum) {
		super(source);
		this.objekt = objekt;
		this.datensatz = datensatz;
		this.datum = datum;
	}

	/**
	 * Gibt den ge&auml;nderten Datensatz zur&uuml;ck.
	 *
	 * @return Menge von Werten
	 */
	public OnlineDatensatz<? extends MesswertDatum> getDatensatz() {
		return datensatz;
	}

	/**
	 * Gibt das aktuelle Datum mit dem Messwerten zur&uuml;ck.
	 *
	 * @return ein Datum mit Messwerten.
	 */
	public MesswertDatum getDatum() {
		return datum;
	}

	/**
	 * Gibt das betreffende Systemobjekt zur&uuml;ck.
	 *
	 * @return Systemobjekt
	 */
	public SystemObjekt getObjekt() {
		return objekt;
	}

	@Override
	public String toString() {
		return "MesswertEvent[objekt=" + objekt //$NON-NLS-1$
				+ ", datensatz=" + datensatz //$NON-NLS-1$
				+ ", datum=" + datum + "]"; //$NON-NLS-1$ //$NON-NLS-2$
	}

}
