/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.param;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.fuzzylib.fuzzylib.LinguistischeVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.FuzzySets;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.PdFuzzySets;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdmsParameterFahrBahnZustandFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdmsParameterGlaetteFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdmsParameterSichtVerhaeltnisseFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdsParameterHelligkeitFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdsParameterNiederschlagsIntensitaetFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdsParameterSichtWeiteFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdsParameterWindGeschwindigkeitMittelWertFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.PdUfdsParameterWindRichtungFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.PdAnteilLkwFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.PdBemessungsDichteFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.PdGeschwindigkeitFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.PdVerkehrsStaerkeStundeFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.PdVerkehrsUnruheFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.PdVerkehrsZustandFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.PdVerkehrsZustandUnruheFuzzy;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.UmfelddatenModellTypen;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsHelligkeit;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsNiederschlagsIntensitaet;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsSichtWeite;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsWindGeschwindigkeitMittelWert;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsWindRichtung;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenMessStelle;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenSensor;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.VerkehrsModellTypen;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.AeusseresStrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.InneresStrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnitt;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittVirtuell;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Tool, welches die Initialparameter aller Fuzzy-Objekte setzt.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class StandardParameterSetzerFuzzy
implements StandardApplication, ClientSenderInterface {

	/** Die Eigenschaft {@code dav}. */
	private static ClientDavInterface dav;

	/**
	 * Gibt die Verbindung der Applikation zum Datenverteiler zur&uuml;ck.
	 *
	 * @return die Datenverteilerverbindung oder {@code null}, wenn
	 *         {@link #initialize(ClientDavInterface)} noch nicht aufgerufen
	 *         wurde.
	 */
	public static ClientDavInterface getDav() {
		return dav;
	}

	/**
	 * Startet die Applikation. Es werden alle Messquerschnitte (normal und
	 * virtuell), Stra&szlig;enteilsegmente, Umfelddatenmessstellen und einige
	 * Umfelddatensensoren parametriert. Mit dem optionalen Parameter
	 * {@code -objekte} kann die Auswahl der Objekte eingeschränkt werden.
	 *
	 * @param args
	 *            neben den Datenverteilerparametern ist {@code -objekte}
	 *            erlaubt.
	 */
	public static void main(final String[] args) {
		StandardApplicationRunner.run(new StandardParameterSetzerFuzzy(), args);
	}

	/** Die Eigenschaft {@code log}. */
	private Debug log;

	/** Die Eigenschaft {@code sendePuffer}. */
	private final List<ResultData> sendePuffer = new ArrayList<>();

	/** Die Eigenschaft {@code objektPids}. */
	private String[] objektPids;

	@Override
	public void dataRequest(final SystemObject object,
			final DataDescription dataDescription, final byte state) {
		if (state == ClientSenderInterface.START_SENDING) {
			synchronized (sendePuffer) {
				final Iterator<ResultData> iterator = sendePuffer.iterator();

				while (iterator.hasNext()) {
					final ResultData rd = iterator.next();
					if (rd.getObject().equals(object) && rd.getDataDescription()
							.equals(dataDescription)) {
						try {
							log.fine("Sende Wissensbasis für", rd.getObject());
							dav.sendData(rd);
							iterator.remove();
						} catch (final DataNotSubscribedException ex) {
							log.error(
									"Wissensbasis konnte als Parameter gesetzt werden",
									new Object[] { rd, ex });
						} catch (final SendSubscriptionNotConfirmed ex) {
							log.error(
									"Wissensbasis konnte als Parameter gesetzt werden",
									new Object[] { rd, ex });
						}
					}
				}
			}
		}
	}

	@Override
	public void initialize(final ClientDavInterface connection)
			throws Exception {
		final ObjektFactory factory = ObjektFactory.getInstanz();
		final List<SystemObjekt> verkehrObj = new ArrayList<>();
		final List<SystemObjekt> umfeldObj = new ArrayList<>();
		long zeitstempel = System.currentTimeMillis();

		dav = connection;

		factory.setVerbindung(dav);
		factory.registerStandardFactories();

		if (objektPids != null) {
			final List<SystemObjekt> objekte = factory
					.bestimmeModellobjekte(objektPids);

			for (final SystemObjekt so : objekte) {
				if (so != null) {
					if ((so instanceof MessQuerschnittAllgemein)
							|| (so instanceof StrassenTeilSegment)) {
						verkehrObj.add(so);
					} else if ((so instanceof UmfeldDatenMessStelle)
							|| (so instanceof UmfeldDatenSensor)) {
						umfeldObj.add(so);
					}
				}
			}
		} else {
			umfeldObj.addAll(factory.bestimmeModellobjekte(
					UmfelddatenModellTypen.UMFELDDATENMESSSTELLE.getPid()));
			umfeldObj.addAll(factory.bestimmeModellobjekte(
					UmfelddatenModellTypen.UDS_HELLIGKEIT.getPid()));
			umfeldObj.addAll(factory.bestimmeModellobjekte(
					UmfelddatenModellTypen.UDS_NIEDERSCHLAGSINTENSITAET
					.getPid()));
			umfeldObj.addAll(factory.bestimmeModellobjekte(
					UmfelddatenModellTypen.UDS_SICHTWEITE.getPid()));
			umfeldObj.addAll(factory.bestimmeModellobjekte(
					UmfelddatenModellTypen.UDS_WINDGESCHWINDIGKEIT.getPid()));
			umfeldObj.addAll(factory.bestimmeModellobjekte(
					UmfelddatenModellTypen.UDS_WINDRICHTUNG.getPid()));

			verkehrObj.addAll(factory.bestimmeModellobjekte(
					VerkehrsModellTypen.MESSQUERSCHNITTALLGEMEIN.getPid()));
			verkehrObj.addAll(factory.bestimmeModellobjekte(
					VerkehrsModellTypen.STRASSENTEILSEGMENT.getPid()));
		}

		parametriereUmfelddatenmodell(umfeldObj);
		parametriereVerkehrsmodell(verkehrObj);

		zeitstempel = System.currentTimeMillis() - zeitstempel;
		log.info("Gesamtparametrierung in " + (zeitstempel / 1000)
				+ " Sekunden.");

		Thread.sleep(10 * Constants.MILLIS_PER_SECOND);
		dav.disconnect(false, "Alles O.k.");
	}

	@Override
	public boolean isRequestSupported(final SystemObject object,
			final DataDescription dataDescription) {
		return true;
	}

	/**
	 * Mit dem optionalen Parameter {@code -objekte} kann die Auswahl der
	 * Objekte eingeschränkt werden.
	 */
	@Override
	public void parseArguments(final ArgumentList argumentList)
			throws Exception {
		log = Debug.getLogger();

		if (argumentList.hasArgument("-objekte")) {
			objektPids = argumentList.fetchArgument("-objekte")
					.asNonEmptyString().split(",");
		}
	}

	/**
	 * Setzt den Parameter für den Lkw-Anteil.
	 *
	 * @param so
	 *            ein {@code MessQuerschnittAllgemein} oder
	 *            {@code Stra&szlig;enteilsegment}.
	 */
	private void parametriereALkw(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze ALkw für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdAnteilLkwFuzzy.class, factory.getLvALkw());

	}

	/**
	 * Setzt den Parameter für den Fahrbahnzustand.
	 *
	 * @param so
	 *            ein Objekt mit dem Fuzzy-Wert Fahrbahnzustand.
	 */
	private void parametriereFahrbahnzustand(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Fahrbahnzustand für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdUfdmsParameterFahrBahnZustandFuzzy.class,
				factory.getLvFahrBahnZustand());
	}

	/**
	 * Setzt den Parameter für den Gl&auml;tte.
	 *
	 * @param so
	 *            ein Objekt mit dem Fuzzy-Wert Gl&auml;tte.
	 */
	private void parametriereGlaette(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Glätte für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdUfdmsParameterGlaetteFuzzy.class,
				factory.getLvGlaette());
	}

	/**
	 * Setzt den Parameter für die Helligkeit.
	 *
	 * @param so
	 *            ein Umfelddatensensor f&uuml;r Helligkeit.
	 */
	private void parametriereHelligkeit(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Helligkeit für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdUfdsParameterHelligkeitFuzzy.class,
				factory.getLvHelligkeit());
	}

	/**
	 * Setzt den Parameter für die Bemessungsdichte.
	 *
	 * @param so
	 *            ein {@code MessQuerschnittAllgemein} oder
	 *            {@code Stra&szlig;enteilsegment}.
	 */
	private void parametriereKB(final SystemObjekt so) {
		final FuzzyParameterFactory factory;
		final LinguistischeVariable lv;

		log.fine("Setze KB für", so);

		factory = FuzzyParameterFactory.getInstanz();
		if (so instanceof MessQuerschnittVirtuell) {
			final MessQuerschnittVirtuell mq = (MessQuerschnittVirtuell) so;

			if (mq.getStrassenSegment() == null) {
				lv = factory.getLvKBSzenario1();
			} else {
				if (mq.getStrassenSegment() instanceof AeusseresStrassenSegment) {
					lv = factory.getLvKBSzenario1();
				} else {
					// MQ muss auf einem inneren Straßensegment liegen
					lv = factory.getLvKBSzenario4();
				}
			}
		} else if (so instanceof MessQuerschnitt) {
			final MessQuerschnitt mq = (MessQuerschnitt) so;

			if (mq.getStrassenSegment() == null) {
				switch (mq.getFahrStreifen().size()) {
				case 2:
					lv = factory.getLvKBSzenario1();
					break;
				case 3:
					lv = factory.getLvKBSzenario2();
					break;
				case 4:
					lv = factory.getLvKBSzenario3();
					break;
				default:
					lv = factory.getLvKBSzenario1();
				}
			} else {
				if (mq.getStrassenSegment() instanceof AeusseresStrassenSegment) {
					switch (mq.getFahrStreifen().size()) {
					case 2:
						lv = factory.getLvKBSzenario1();
						break;
					case 3:
						lv = factory.getLvKBSzenario2();
						break;
					case 4:
						lv = factory.getLvKBSzenario3();
						break;
					default:
						lv = factory.getLvKBSzenario1();
					}
				} else {
					// MQ muss auf einem inneren Straßensegment liegen
					switch (mq.getFahrStreifen().size()) {
					case 2:
						lv = factory.getLvKBSzenario4();
						break;
					case 3:
						lv = factory.getLvKBSzenario5();
						break;
					case 4:
						lv = factory.getLvKBSzenario6();
						break;
					default:
						lv = factory.getLvKBSzenario4();
					}
				}
			}
		} else if (so instanceof StrassenTeilSegment) {
			final StrassenTeilSegment sts = (StrassenTeilSegment) so;

			if ((sts.getStrassenSegment().size() > 0)
					&& (sts.getStrassenSegment()
							.get(0) instanceof InneresStrassenSegment)) {
				// MQ liegt auf einem inneren Straßensegment
				switch (sts.getAnzahlFahrstreifen()) {
				case 2:
					lv = factory.getLvKBSzenario4();
					break;
				case 3:
					lv = factory.getLvKBSzenario5();
					break;
				case 4:
					lv = factory.getLvKBSzenario6();
					break;
				default:
					lv = factory.getLvKBSzenario1();
				}
			} else {
				// MQ liegt auf einem äußeren Straßensegment oder auf keinem
				switch (sts.getAnzahlFahrstreifen()) {
				case 2:
					lv = factory.getLvKBSzenario1();
					break;
				case 3:
					lv = factory.getLvKBSzenario2();
					break;
				case 4:
					lv = factory.getLvKBSzenario3();
					break;
				default:
					lv = factory.getLvKBSzenario1();
				}
			}
		} else {
			log.error(
					"Objekt ist weder vom Typ MessQuerschnitt, MessQuerschnittVirtuell noch StraßenTeilSegment",
					so);
			return;
		}

		assert lv.getName().equals("KB");
		sendeFuzzySets(so, PdBemessungsDichteFuzzy.class, lv);
	}

	/**
	 * Setzt den Parameter für den Niederschlag.
	 *
	 * @param so
	 *            ein Umfelddatensensor f&uuml;r Niederschlag.
	 */
	private void parametriereNiederschlag(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Niederschlag für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdUfdsParameterNiederschlagsIntensitaetFuzzy.class,
				factory.getLvNiederschlagsIntensitaet());
	}

	/**
	 * Setzt den Parameter für die Verkehrsst&auml;rke.
	 *
	 * @param so
	 *            ein {@code MessQuerschnittAllgemein} oder
	 *            {@code Stra&szlig;enteilsegment}.
	 */
	private void parametriereQKfz(final SystemObjekt so) {
		final FuzzyParameterFactory factory;
		final LinguistischeVariable lv;

		log.fine("Setze QKfz für", so);

		factory = FuzzyParameterFactory.getInstanz();
		if (so instanceof MessQuerschnittVirtuell) {
			lv = factory.getLvQKfzSzenario1();
		} else if (so instanceof MessQuerschnitt) {
			final MessQuerschnitt mq = (MessQuerschnitt) so;

			switch (mq.getFahrStreifen().size()) {
			case 2:
				lv = factory.getLvQKfzSzenario1();
				break;
			case 3:
				lv = factory.getLvQKfzSzenario2();
				break;
			case 4:
				lv = factory.getLvQKfzSzenario3();
				break;
			default:
				lv = factory.getLvQKfzSzenario1();
			}
		} else if (so instanceof StrassenTeilSegment) {
			final StrassenTeilSegment sts = (StrassenTeilSegment) so;

			switch (sts.getAnzahlFahrstreifen()) {
			case 2:
				lv = factory.getLvQKfzSzenario1();
				break;
			case 3:
				lv = factory.getLvQKfzSzenario2();
				break;
			case 4:
				lv = factory.getLvQKfzSzenario3();
				break;
			default:
				lv = factory.getLvQKfzSzenario1();
			}
		} else {
			log.error(
					"Objekt ist weder vom Typ MessQuerschnitt, MessQuerschnittVirtuell noch StraßenTeilSegment",
					so);
			return;
		}

		assert lv.getName().equals("QKfz");
		sendeFuzzySets(so, PdVerkehrsStaerkeStundeFuzzy.class, lv);
	}

	/**
	 * Setzt den Parameter für den Sichtverh&auml;ltnisse.
	 *
	 * @param so
	 *            ein Objekt mit dem Fuzzy-Wert Sichtverh&auml;ltnisse.
	 */
	private void parametriereSichtverhaeltnisse(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Sichtverhältnisse für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdUfdmsParameterSichtVerhaeltnisseFuzzy.class,
				factory.getLvSichtVerhaeltnisse());
	}

	/**
	 * Setzt den Parameter für die Sichtweite.
	 *
	 * @param so
	 *            ein Umfelddatensensor f&uuml;r Sichtweite.
	 */
	private void parametriereSichtWeite(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Sichtweite für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdUfdsParameterSichtWeiteFuzzy.class,
				factory.getLvSichtWeite());
	}

	/**
	 * Setzt den Parameter für die Standardabweichung von der Geschwindigkeit.
	 *
	 * @param so
	 *            ein {@code MessQuerschnittAllgemein} oder
	 *            {@code Stra&szlig;enteilsegment}.
	 */
	private void parametriereSKfz(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze SKfz für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdVerkehrsUnruheFuzzy.class, factory.getLvSKfz());
	}

	/**
	 * Parametriert die Objekte des Umfeldatenmodells.
	 *
	 * @param objekte
	 *            die Liste der zu parametrierenden Umfelddatenobjekte.
	 */
	private void parametriereUmfelddatenmodell(
			final List<SystemObjekt> objekte) {
		long zeitstempel = System.currentTimeMillis();

		log.info("Beginne Parametrierung des Umfelddatenmodells für "
				+ objekte.size() + " Objekte.");

		for (int i = 0; i < objekte.size(); i++) {
			final SystemObjekt so = objekte.get(i);
			ResultData wb;

			log.info("Parametriere " + (i + 1) + " von " + objekte.size(), so);

			if (so instanceof UmfeldDatenMessStelle) {
				parametriereFahrbahnzustand(so);
				parametriereGlaette(so);
				parametriereSichtverhaeltnisse(so);

				wb = FuzzyParameterFactory.erzeugeWissensbasisUmfelddatenModell(
						so.getSystemObject());
				setzeWissensbasis(wb);
			} else if (so instanceof UfdsHelligkeit) {
				parametriereHelligkeit(so);
			} else if (so instanceof UfdsNiederschlagsIntensitaet) {
				parametriereNiederschlag(so);
			} else if (so instanceof UfdsSichtWeite) {
				parametriereSichtWeite(so);
			} else if (so instanceof UfdsWindGeschwindigkeitMittelWert) {
				parametriereWindgeschwindigkeit(so);
			} else if (so instanceof UfdsWindRichtung) {
				parametriereWindrichtung(so);
			}
		}

		zeitstempel = System.currentTimeMillis() - zeitstempel;
		log.info("Parametrierung des Umfelddatenmodells in "
				+ (zeitstempel / 1000) + " Sekunden.");
	}

	/**
	 * Parametriert die Objekte des Verkehrsmodells.
	 *
	 * @param objekte
	 *            die Liste der zu parametrierenden Verkehrsobjekte.
	 */
	private void parametriereVerkehrsmodell(final List<SystemObjekt> objekte) {
		long zeitstempel = System.currentTimeMillis();

		log.info("Beginne Parametrierung des Verkehrsmodells mit "
				+ objekte.size() + " Objekten.");
		for (int i = 0; i < objekte.size(); i++) {
			final SystemObjekt so = objekte.get(i);
			ResultData wb;

			log.info("Parametriere " + (i + 1) + " von " + objekte.size(), so);

			parametriereALkw(so);
			parametriereKB(so);
			parametriereQKfz(so);
			parametriereSKfz(so);
			parametriereVKfz(so);
			parametriereVerkehrsZustand(so);
			parametriereVerkehrsZustandUnruhe(so);

			wb = FuzzyParameterFactory
					.erzeugeWissensbasisVerkehrsmodell(so.getSystemObject());
			setzeWissensbasis(wb);
		}

		zeitstempel = System.currentTimeMillis() - zeitstempel;
		log.info("Parametrierung des Verkehrsmodells in " + (zeitstempel / 1000)
				+ " Sekunden beendet.");
	}

	/**
	 * Setzt den Parameter für den Verkehrszustand.
	 *
	 * @param so
	 *            ein {@code MessQuerschnittAllgemein} oder
	 *            {@code Stra&szlig;enteilsegment}.
	 */
	private void parametriereVerkehrsZustand(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Verkehrszustand für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdVerkehrsZustandFuzzy.class,
				factory.getLvVerkehrsZustand());
	}

	/**
	 * Setzt den Parameter für den Unruhe im Verkehr.
	 *
	 * @param so
	 *            ein {@code MessQuerschnittAllgemein} oder
	 *            {@code Stra&szlig;enteilsegment}.
	 */
	private void parametriereVerkehrsZustandUnruhe(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Unruhe im Verkehr für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdVerkehrsZustandUnruheFuzzy.class,
				factory.getLvVerkehrsZustandUnruhe());
	}

	/**
	 * Setzt den Parameter für die Geschwindigkeit.
	 *
	 * @param so
	 *            ein {@code MessQuerschnittAllgemein} oder
	 *            {@code Stra&szlig;enteilsegment}.
	 */
	private void parametriereVKfz(final SystemObjekt so) {
		final FuzzyParameterFactory factory;
		final LinguistischeVariable lv;

		log.fine("Setze VKfz für", so);

		factory = FuzzyParameterFactory.getInstanz();

		if (so instanceof MessQuerschnittVirtuell) {
			final MessQuerschnittVirtuell mq = (MessQuerschnittVirtuell) so;

			if (mq.getStrassenSegment() == null) {
				lv = factory.getLvVKfzSzenario1();
			} else {
				if (mq.getStrassenSegment() instanceof AeusseresStrassenSegment) {
					lv = factory.getLvVKfzSzenario1();
				} else {
					// MQ muss auf einem inneren Straßensegment liegen
					lv = factory.getLvVKfzSzenario4();
				}
			}
		} else if (so instanceof MessQuerschnitt) {
			final MessQuerschnitt mq = (MessQuerschnitt) so;

			if (mq.getStrassenSegment() == null) {
				switch (mq.getFahrStreifen().size()) {
				case 2:
					lv = factory.getLvVKfzSzenario1();
					break;
				case 3:
					lv = factory.getLvVKfzSzenario2();
					break;
				case 4:
					lv = factory.getLvVKfzSzenario3();
					break;
				default:
					lv = factory.getLvVKfzSzenario1();
				}
			} else {
				if (mq.getStrassenSegment() instanceof AeusseresStrassenSegment) {
					switch (mq.getFahrStreifen().size()) {
					case 2:
						lv = factory.getLvVKfzSzenario1();
						break;
					case 3:
						lv = factory.getLvVKfzSzenario2();
						break;
					case 4:
						lv = factory.getLvVKfzSzenario3();
						break;
					default:
						lv = factory.getLvVKfzSzenario1();
					}
				} else {
					// MQ muss auf einem inneren Straßensegment liegen
					switch (mq.getFahrStreifen().size()) {
					case 2:
						lv = factory.getLvVKfzSzenario4();
						break;
					case 3:
						lv = factory.getLvVKfzSzenario5();
						break;
					case 4:
						lv = factory.getLvVKfzSzenario6();
						break;
					default:
						lv = factory.getLvVKfzSzenario4();
					}
				}
			}
		} else if (so instanceof StrassenTeilSegment) {
			final StrassenTeilSegment sts = (StrassenTeilSegment) so;

			if (sts.getStrassenSegment() instanceof AeusseresStrassenSegment) {
				switch (sts.getAnzahlFahrstreifen()) {
				case 2:
					lv = factory.getLvVKfzSzenario1();
					break;
				case 3:
					lv = factory.getLvVKfzSzenario2();
					break;
				case 4:
					lv = factory.getLvVKfzSzenario3();
					break;
				default:
					lv = factory.getLvVKfzSzenario1();
				}
			} else {
				// MQ muss auf einem inneren Straßensegment liegen
				switch (sts.getAnzahlFahrstreifen()) {
				case 2:
					lv = factory.getLvVKfzSzenario4();
					break;
				case 3:
					lv = factory.getLvVKfzSzenario5();
					break;
				case 4:
					lv = factory.getLvVKfzSzenario6();
					break;
				default:
					lv = factory.getLvVKfzSzenario1();
				}
			}
		} else {
			log.error(
					"Objekt ist weder vom Typ MessQuerschnitt, MessQuerschnittVirtuell noch StraßenTeilSegment",
					so);
			return;
		}

		assert lv.getName().equals("VKfz");
		sendeFuzzySets(so, PdGeschwindigkeitFuzzy.class, lv);
	}

	/**
	 * Setzt den Parameter für den Windgeschwindigkeit.
	 *
	 * @param so
	 *            ein Umfelddatensensor f&uuml;r Windgeschwindigkeit.
	 */
	private void parametriereWindgeschwindigkeit(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Windgeschwindigkeit für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so,
				PdUfdsParameterWindGeschwindigkeitMittelWertFuzzy.class,
				factory.getLvWindGeschwindigkeitMittelWert());

	}

	/**
	 * Setzt den Parameter für den Windrichtung.
	 *
	 * @param so
	 *            ein Umfelddatensensor f&uuml;r Windrichtung.
	 */
	private void parametriereWindrichtung(final SystemObjekt so) {
		FuzzyParameterFactory factory;

		log.fine("Setze Windrichtung für", so);

		factory = FuzzyParameterFactory.getInstanz();
		sendeFuzzySets(so, PdUfdsParameterWindRichtungFuzzy.class,
				factory.getLvWindRichtung());
	}

	/**
	 * Versucht einen Parameter zu senden. Der Parameter wird zum Senden
	 * angemeldet. Wenn die Sendesteuerung positiv ist, wird er gesendet. Zum
	 * Schluss wird die Sendeanmeldung wieder aufgehoben.
	 *
	 * @param so
	 *            das Systemobjekt.
	 * @param fuzzySets
	 *            der Typ des Parameters.
	 * @param lv
	 *            die Daten des Parameters.
	 */
	private static void sendeFuzzySets(final SystemObjekt so,
			final Class<? extends PdFuzzySets> fuzzySets,
			final LinguistischeVariable lv) {
		FuzzySets fs;
		PdFuzzySets param;

		param = so.getParameterDatensatz(fuzzySets);
		fs = param.erzeugeDatum();
		fs.setLinguistischeVariable(lv);
		try {
			param.anmeldenSender();
			param.sendeDaten(fs, 60 * 1000);
			param.abmeldenSender();
		} catch (final DatensendeException ex) {
			ex.printStackTrace();
			System.exit(-1);
		} catch (final AnmeldeException ex) {
			ex.printStackTrace();
			System.exit(-1);
		}
	}

	/**
	 * Setzt den Parameter mit der Wissensbasis.
	 *
	 * @param daten
	 *            die zu versenden Daten.
	 */
	private void setzeWissensbasis(final ResultData daten) {
		log.fine("Puffere Wissensbasis für", daten.getObject());

		try {
			dav.subscribeSender(this, daten.getObject(),
					daten.getDataDescription(), SenderRole.sender());
			synchronized (sendePuffer) {
				sendePuffer.add(daten);
			}
		} catch (final OneSubscriptionPerSendData ex) {
			log.error(
					"Sendeanmeldung für Parameter konnte nicht durchgeführt werden",
					daten);
		}
	}

}
