/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.param;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzySet;
import de.bsvrz.iav.fuzzylib.fuzzylib.LinguistischeVariable;

/**
 * Factory, die Prototypen der Fuzzy-Parameter (linguistische Variablen und
 * Wissensbasen) liefert.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public final class FuzzyParameterFactory {

	/** Attribut. */
	private static final String ALKW = "ALkw";

	/** Attribut. */
	private static final String QKFZ = "QKfz";

	/** Attribut. */
	private static final String SKFZ = "SKfz";

	/** Attribut. */
	private static final String VKFZ = "VKfz";

	/** Attribut. */
	private static final String KB = "KB";

	/** Attribut. */
	private static final String VERKEHRSZUSTAND = "VerkehrsZustand";

	/** Attribut. */
	private static final String VERKEHRSZUSTAND_UNRUHE = "VerkehrsZustandUnruhe";

	/** Attribut. */
	private static final String HELLIGKEIT = "Helligkeit";

	/** Attribut. */
	private static final String NI = "NiederschlagsIntensität";

	/** Attribut. */
	private static final String SICHTWEITE = "SichtWeite";

	/** Attribut. */
	private static final String WG = "WindGeschwindigkeitMittelWert";

	/** Attribut. */
	private static final String WR = "WindRichtung";

	/** Attribut. */
	private static final String GLAETTE = "Glätte";

	/** Attribut. */
	private static final String FAHRBAHNZUSTAND = "FahrBahnZustand";

	/** Attribut. */
	private static final String SICHT = "SichtVerhältnisse";

	/** Fuzzy-Term. */
	private static final String SEHR_GERING = "Sehr gering";

	/** Fuzzy-Term. */
	private static final String GERING = "Gering";

	/** Fuzzy-Term. */
	private static final String SEHR_NIEDRIG = "Sehr niedrig";

	/** Fuzzy-Term. */
	private static final String NIEDRIG = "Niedrig";

	/** Fuzzy-Term. */
	private static final String MITTEL = "Mittel";

	/** Fuzzy-Term. */
	private static final String HOCH = "Hoch";

	/** Fuzzy-Term. */
	private static final String SEHR_HOCH = "Sehr hoch";

	/** Fuzzy-Term. */
	private static final String NICHT_SEHR_HOCH = "Nicht sehr hoch";

	/** Fuzzy-Term. */
	private static final String JA = "Ja";

	/** Fuzzy-Term. */
	private static final String NEIN = "Nein";

	/** Fuzzy-Term. */
	private static final String GESTAUT = "Gestaut";

	/** Fuzzy-Term. */
	private static final String DICHT = "Dicht";

	/** Fuzzy-Term. */
	private static final String ZAEH = "Zähfließend";

	/** Fuzzy-Term. */
	private static final String FREI = "Frei";

	/** Fuzzy-Term. */
	private static final String UNRUHE = "Unruhe im Verkehr";

	/** Fuzzy-Term. */
	private static final String KEINE_UNRUHE = "Keine Unruhe";

	/** Fuzzy-Term. */
	private static final String KEIN_NI = "Kein Niederschlag";

	/** Fuzzy-Term. */
	private static final String SCHWACHER_NI = "Schwacher Niederschlag";

	/** Fuzzy-Term. */
	private static final String MITTLERER_NI = "Mittlerer Niederschlag";

	/** Fuzzy-Term. */
	private static final String STARKER_NI = "Starker Niederschlag";

	/** Fuzzy-Term. */
	private static final String SEHR_STARKER_NI = "Sehr starker Niederschlag";

	/** Fuzzy-Term. */
	private static final String N = "Nord";

	/** Fuzzy-Term. */
	private static final String NO = "Nordost";

	/** Fuzzy-Term. */
	private static final String O = "Ost";

	/** Fuzzy-Term. */
	private static final String SO = "Südost";

	/** Fuzzy-Term. */
	private static final String S = "Süd";

	/** Fuzzy-Term. */
	private static final String SW = "Südwest";

	/** Fuzzy-Term. */
	private static final String W = "West";

	/** Fuzzy-Term. */
	private static final String NW = "Nordwest";

	/** Fuzzy-Term. */
	private static final String KEIN_WIND = "Kein Wind";

	/** Fuzzy-Term. */
	private static final String SCHWACHER_WIND = "Schwacher Wind";

	/** Fuzzy-Term. */
	private static final String MITTLERER_WIND = "Mittlerer Wind";

	/** Fuzzy-Term. */
	private static final String STARKER_WIND = "Starker Wind";

	/** Fuzzy-Term. */
	private static final String DUNKEL = "Dunkel";

	/** Fuzzy-Term. */
	private static final String DAEMMRIG = "Dämmrig";

	/** Fuzzy-Term. */
	private static final String HELL = "Hell";

	/** Fuzzy-Term. */
	private static final String SEHR_HELL = "Sehr hell";

	/** Fuzzy-Term. */
	private static final String TROCKEN = "Trocken";

	/** Fuzzy-Term. */
	private static final String NASS = "Nass";

	/** Fuzzy-Term. */
	private static final String GLATT = "Glatt";

	/** Fuzzy-Term. */
	private static final String GUT = "Gut";

	/** Fuzzy-Term. */
	private static final String LEICHT_EINGESCHRAENKT = "Leicht eingeschränkt";

	/** Fuzzy-Term. */
	private static final String EINGESCHRAENKT = "Eingeschränkt";

	/** Fuzzy-Term. */
	private static final String UNZUREICHEND = "Unzureichend";

	/** Fuzzy-Term. */
	private static final String SICHT_BIS_50M = "Sicht bis 50 m";

	/** Fuzzy-Term. */
	private static final String SICHT_BIS_80M = "Sicht bis 80 m";

	/** Fuzzy-Term. */
	private static final String SICHT_BIS_120M = "Sicht bis 120 m";

	/** Fuzzy-Term. */
	private static final String SICHT_BIS_250M = "Sicht bis 250 m";

	/** Fuzzy-Term. */
	private static final String SICHT_BIS_400M = "Sicht bis 400 m";

	/** Fuzzy-Term. */
	private static final String SICHT_UEBER_400M = "Sicht über 400 m";

	/** PID. */
	private static final String ATG_FUZZY_REGELBASIS = "atg.fuzzyRegelbasis";

	/** Attribut. */
	private static final String FELD_REGELN = "Regeln";

	/** Attribut. */
	private static final String FELD_PRAEMISSE = "Prämisse";

	/** Attribut. */
	private static final String ATL_KONKLUSION = "Konklusion";

	/** Attribut. */
	private static final String ATT_VARIABLE = "Variable";

	/** Attribut. */
	private static final String ATT_TERM = "Term";

	/** Attribut. */
	private static final String FELD_TERM = "Term";

	/** Attribut. */
	private static final String ATT_ZYKLUS = "Zyklus";

	/** Die einzige Instanz der Klasse. */
	private static FuzzyParameterFactory singleton;

	/**
	 * Erzeugt die Wissensbasis des Verkehrsmodells.
	 *
	 * @param so
	 *            das Systemobject, dessen Wissensbasis gesucht wird.
	 * @return ein Datum mit der Wissensbasis.
	 */
	public static ResultData erzeugeWissensbasisUmfelddatenModell(
			final SystemObject so) {
		final ClientDavInterface dav = StandardParameterSetzerFuzzy.getDav();
		final DataModel modell = dav.getDataModel();
		final AttributeGroup atg = modell
				.getAttributeGroup(ATG_FUZZY_REGELBASIS);
		final Aspect asp = modell.getAspect("asp.parameterVorgabe");
		final Data datum = StandardParameterSetzerFuzzy.getDav()
				.createData(atg);
		Array regeln, praemissen, terme;
		Data regel, praemisse, konklusion, term;

		regeln = datum.getArray(FELD_REGELN);
		regeln.setLength(25);

		// FahrBahnZustand.Trocken <- FahrBahnGlätte.Nein UND
		// NiederschlagsIntensität.Kein
		// Niederschlag
		regel = regeln.getItem(0);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NEIN);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(KEIN_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(TROCKEN);

		// FahrBahnZustand.Nass <- FahrBahnGlätte.Nein UND
		// NiederschlagsIntensität.Schwacher
		// Niederschlag
		regel = regeln.getItem(1);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NEIN);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SCHWACHER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(NASS);

		// FahrBahnZustand.Nass <- FahrBahnGlätte.Nein UND
		// NiederschlagsIntensität.Mittlerer
		// Niederschlag
		regel = regeln.getItem(2);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NEIN);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTLERER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(NASS);

		// FahrBahnZustand.Nass <- FahrBahnGlätte.Nein UND
		// NiederschlagsIntensität.Starker
		// Niederschlag
		regel = regeln.getItem(3);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NEIN);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(STARKER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(NASS);

		// FahrBahnZustand.Nass <- FahrBahnGlätte.Nein UND
		// NiederschlagsIntensität.Sehr starker
		// Niederschlag
		regel = regeln.getItem(4);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NEIN);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_STARKER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(NASS);

		// FahrBahnZustand.Glatt <- FahrBahnGlätte.Ja UND Ni.Kein Niederschlag
		regel = regeln.getItem(5);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(JA);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(KEIN_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(GLATT);

		// FahrBahnZustand.Glatt <- FahrBahnGlätte.Ja UND
		// NiederschlagsIntensität.Schwacher
		// Niederschlag
		regel = regeln.getItem(6);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(JA);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SCHWACHER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(GLATT);

		// FahrBahnZustand.Glatt <- FahrBahnGlätte.Ja UND
		// NiederschlagsIntensität.Mittlerer
		// Niederschlag
		regel = regeln.getItem(7);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(JA);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTLERER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(GLATT);

		// FahrBahnZustand.Glatt <- FahrBahnGlätte.Ja UND
		// NiederschlagsIntensität.Starker
		// Niederschlag
		regel = regeln.getItem(8);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(GLAETTE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(JA);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(STARKER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(FAHRBAHNZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(GLATT);

		// SichtVerhältnisse.Gut <- Helligkeit.Sehr hell UND (SichtWeite.bis
		// 250m ODER SichtWeite.bis 400m ODER SichtWeite.über 400m) UND
		// NiederschlagsIntensität.Kein Niederschlag
		regel = regeln.getItem(9);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_HELL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_400M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_UEBER_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(KEIN_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(GUT);

		// SichtVerhältnisse.Gut <- Helligkeit.Sehr hell UND (SichtWeite.bis
		// 250m ODER SichtWeite.bis 400m ODER SichtWeite.über 400m) UND
		// NiederschlagsIntensität.Schwacher Niederschlag
		regel = regeln.getItem(10);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_HELL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_400M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_UEBER_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SCHWACHER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(GUT);

		// SichtVerhältnisse.Leicht eingeschränkt <- Helligkeit.Hell UND
		// (SichtWeite.bis 250m ODER SichtWeite.bis 400m ODER SichtWeite.über
		// 400m) UND NiederschlagsIntensität.Kein Niederschlag
		regel = regeln.getItem(11);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HELL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_400M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_UEBER_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(KEIN_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(LEICHT_EINGESCHRAENKT);

		// SichtVerhältnisse.Leicht eingeschränkt <- Helligkeit.Hell UND
		// (SichtWeite.bis 250m ODER SichtWeite.bis 400m ODER SichtWeite.über
		// 400m) UND NiederschlagsIntensität.Schwacher Niederschlag
		regel = regeln.getItem(12);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HELL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_400M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_UEBER_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SCHWACHER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(LEICHT_EINGESCHRAENKT);

		// SichtVerhältnisse.Leicht eingeschränkt <- Helligkeit.Dämmrig UND
		// (SichtWeite.bis 250m ODER SichtWeite.bis 400m ODER SichtWeite.über
		// 400m) UND NiederschlagsIntensität.Kein Niederschlag
		regel = regeln.getItem(13);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DAEMMRIG);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_400M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_UEBER_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(KEIN_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(LEICHT_EINGESCHRAENKT);

		// SichtVerhältnisse.Leicht eingeschränkt <- Helligkeit.Dämmrig UND
		// (SichtWeite.bis 120m ODER SichtWeite.bis 250m ODER SichtWeite.bis
		// 400m) UND NiederschlagsIntensität.Schwacher Niederschlag
		regel = regeln.getItem(14);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DAEMMRIG);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_120M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SCHWACHER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(LEICHT_EINGESCHRAENKT);

		// SichtVerhältnisse.Eingeschränkt <- Helligkeit.Sehr hell UND
		// (SichtWeite.bis 120m ODER SichtWeite.bis 250m ODER SichtWeite.bis
		// 400m) UND NiederschlagsIntensität.Mittlerer Niederschlag
		regel = regeln.getItem(15);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_HELL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_120M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTLERER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(EINGESCHRAENKT);

		// SichtVerhältnisse.Eingeschränkt <- Helligkeit.Hell UND
		// (SichtWeite.bis 120m ODER SichtWeite.bis 250m ODER SichtWeite.bis
		// 400m) UND NiederschlagsIntensität.Mittlerer Niederschlag
		regel = regeln.getItem(16);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HELL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_120M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_250M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_400M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTLERER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(EINGESCHRAENKT);

		// SichtVerhältnisse.Eingeschränkt <- Helligkeit.Dämmrig UND
		// (SichtWeite.bis 80m ODER SichtWeite.bis 120m ODER SichtWeite.bis
		// 250m) UND NiederschlagsIntensität.Mittlerer Niederschlag
		regel = regeln.getItem(17);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DAEMMRIG);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_120M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_250M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTLERER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(EINGESCHRAENKT);

		// SichtVerhältnisse.Eingeschränkt <- Helligkeit.Dämmrig UND
		// (SichtWeite.bis 50m ODER SichtWeite.bis 80m ODER SichtWeite.bis
		// 120m) UND NiederschlagsIntensität.Starker Niederschlag
		regel = regeln.getItem(18);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DAEMMRIG);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_50M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_120M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(STARKER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(EINGESCHRAENKT);

		// SichtVerhältnisse.Eingeschränkt <- Helligkeit.Dämmrig UND
		// (SichtWeite.bis 50m ODER SichtWeite.bis 80m ODER SichtWeite.bis
		// 120m) UND NiederschlagsIntensität.Sehr starker Niederschlag
		regel = regeln.getItem(19);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DAEMMRIG);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_50M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_120M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_STARKER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(EINGESCHRAENKT);

		// SichtVerhältnisse.Eingeschränkt <- Helligkeit.Dunkel UND
		// (SichtWeite.bis 80m ODER SichtWeite.bis 120m ODER SichtWeite.bis
		// 250m) UND NiederschlagsIntensität.Kein Niederschlag
		regel = regeln.getItem(20);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DUNKEL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_120M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_250M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(KEIN_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(EINGESCHRAENKT);

		// SichtVerhältnisse.Eingeschränkt <- Helligkeit.Dunkel UND
		// (SichtWeite.bis 80m ODER SichtWeite.bis 120m ODER SichtWeite.bis
		// 250m) UND NiederschlagsIntensität.Schwacher Niederschlag
		regel = regeln.getItem(21);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DUNKEL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_120M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_250M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SCHWACHER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(EINGESCHRAENKT);

		// SichtVerhältnisse.Unzureichend <- Helligkeit.Dunkel UND
		// (SichtWeite.bis 50m ODER SichtWeite.bis 80m ODER SichtWeite.bis
		// 120m) UND NiederschlagsIntensität.Mittlerer Niederschlag
		regel = regeln.getItem(22);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DUNKEL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_50M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_120M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTLERER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(UNZUREICHEND);

		// SichtVerhältnisse.Unzureichend <- Helligkeit.Dunkel UND
		// (SichtWeite.bis 50m ODER SichtWeite.bis 80m ODER SichtWeite.bis
		// 120m) UND NiederschlagsIntensität.Starker Niederschlag
		regel = regeln.getItem(23);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DUNKEL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_50M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_120M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(STARKER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(UNZUREICHEND);

		// SichtVerhältnisse.Unzureichend <- Helligkeit.Dunkel UND
		// (SichtWeite.bis 50m ODER SichtWeite.bis 80m ODER SichtWeite.bis
		// 120m) UND NiederschlagsIntensität.Sehr starker Niederschlag
		regel = regeln.getItem(24);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(3);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(HELLIGKEIT);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(DUNKEL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SICHTWEITE);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(3);
		term = terme.getItem(0);
		term.asTextValue().setText(SICHT_BIS_50M);
		term = terme.getItem(1);
		term.asTextValue().setText(SICHT_BIS_80M);
		term = terme.getItem(2);
		term.asTextValue().setText(SICHT_BIS_120M);
		praemisse = praemissen.getItem(2);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(NI);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_STARKER_NI);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(SICHT);
		konklusion.getTextValue(ATT_TERM).setText(UNZUREICHEND);

		return new ResultData(so, new DataDescription(atg, asp), dav.getTime(),
				datum);
	}

	/**
	 * Erzeugt die Wissensbasis des Verkehrsmodells.
	 *
	 * @param so
	 *            das Systemobject, dessen Wissensbasis gesucht wird.
	 * @return ein Datum mit der Wissensbasis.
	 */
	public static ResultData erzeugeWissensbasisVerkehrsmodell(
			final SystemObject so) {
		final ClientDavInterface dav = StandardParameterSetzerFuzzy.getDav();
		final DataModel modell = dav.getDataModel();
		final AttributeGroup atg = modell
				.getAttributeGroup(ATG_FUZZY_REGELBASIS);
		final Aspect asp = modell.getAspect("asp.parameterVorgabe");
		final Data datum = dav.createData(atg);
		Array regeln, praemissen, terme;
		Data regel, praemisse, konklusion, term;

		regeln = datum.getArray(FELD_REGELN);
		regeln.setLength(11);

		// Verkehrszustand.Gestaut <- VKfz(0).Niedrig UND VKfz(-1).Niedrig
		regel = regeln.getItem(0);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(VKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NIEDRIG);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(VKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(1);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NIEDRIG);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(GESTAUT);

		// Verkehrszustand.Gestaut <- KB.Sehr hoch
		regel = regeln.getItem(1);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(1);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(KB);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_HOCH);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(GESTAUT);

		// Verkehrszustand.Zähfließend <- VKfz.Mittel UND KB.Nicht sehr hoch
		regel = regeln.getItem(2);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(VKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTEL);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(KB);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NICHT_SEHR_HOCH);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(ZAEH);

		// Verkehrszustand.Dicht <- VKfz.Hoch UND KB.hoch
		regel = regeln.getItem(3);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(VKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(KB);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(DICHT);

		// Verkehrszustand.Frei <- VKfz.Hoch UND KB.Niedrig
		regel = regeln.getItem(4);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(VKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(KB);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NIEDRIG);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND);
		konklusion.getTextValue(ATT_TERM).setText(FREI);

		// Unruhe im Verkehrsfluss.Unruhe im Verkehr <- SKfz.Hoch UND QKfz.Hoch
		regel = regeln.getItem(5);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(QKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND_UNRUHE);
		konklusion.getTextValue(ATT_TERM).setText(UNRUHE);

		// Unruhe im Verkehrsfluss.Unruhe im Verkehr <- SKfz.Hoch UND QKfz.Sehr
		// hoch
		regel = regeln.getItem(6);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(QKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_HOCH);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND_UNRUHE);
		konklusion.getTextValue(ATT_TERM).setText(UNRUHE);

		// Unruhe im Verkehrsfluss.keine Unruhe <- SKfz.Hoch UND QKfz.Mittel
		regel = regeln.getItem(7);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(QKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(MITTEL);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND_UNRUHE);
		konklusion.getTextValue(ATT_TERM).setText(KEINE_UNRUHE);

		// Unruhe im Verkehrsfluss.keine Unruhe <- SKfz.Hoch UND QKfz.Niedrig
		regel = regeln.getItem(8);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(QKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NIEDRIG);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND_UNRUHE);
		konklusion.getTextValue(ATT_TERM).setText(KEINE_UNRUHE);

		// Unruhe im Verkehrsfluss.keine Unruhe <- SKfz.Hoch UND QKfz.Sehr
		// Niedrig
		regel = regeln.getItem(9);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(2);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(HOCH);
		praemisse = praemissen.getItem(1);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(QKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(SEHR_NIEDRIG);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND_UNRUHE);
		konklusion.getTextValue(ATT_TERM).setText(KEINE_UNRUHE);

		// Unruhe im Verkehrsfluss.keine Unruhe <- SKfz.Niedrig
		regel = regeln.getItem(10);
		regel.getScaledValue("Sicherheitsgrad").set(1.0);
		praemissen = regel.getArray(FELD_PRAEMISSE);
		praemissen.setLength(1);
		praemisse = praemissen.getItem(0);
		praemisse.getItem(ATT_VARIABLE).asTextValue().setText(SKFZ);
		praemisse.getItem(ATT_ZYKLUS).asUnscaledValue().set(0);
		terme = praemisse.getArray(FELD_TERM);
		terme.setLength(1);
		term = terme.getItem(0);
		term.asTextValue().setText(NIEDRIG);
		konklusion = regel.getItem(ATL_KONKLUSION);
		konklusion.getTextValue(ATT_VARIABLE).setText(VERKEHRSZUSTAND_UNRUHE);
		konklusion.getTextValue(ATT_TERM).setText(KEINE_UNRUHE);

		return new ResultData(so, new DataDescription(atg, asp), dav.getTime(),
				datum);
	}

	/**
	 * Gibt die Factory als Singleton zur&uuml;ck.
	 *
	 * @return die einzige Instanz der Factory.
	 */
	public static FuzzyParameterFactory getInstanz() {
		if (singleton == null) {
			singleton = new FuzzyParameterFactory();
		}
		return singleton;
	}

	/**
	 * Lkw-Anteil (ALkw in Prozent).
	 */
	private LinguistischeVariable lvALkw;

	/**
	 * Bemessungsdicht (KB in Fahrzeuge/km), Szenario I: 2-streifiger
	 * Messquerschnitt auf freier Strecke.
	 */
	private LinguistischeVariable lvKBSzenario1;

	/**
	 * Bemessungsdichte (KB in Fahrzeuge/km), Szenario II: 3-streifiger
	 * Messquerschnitt auf freier Strecke.
	 */
	private LinguistischeVariable lvKBSzenario2;

	/**
	 * Bemessungsdichte (KB in Fahrzeuge/km), Szenario III: 4-streifiger
	 * Messquerschnitt auf freier Strecke.
	 */
	private LinguistischeVariable lvKBSzenario3;

	/**
	 * Bemessungsdichte (KB in Fahrzeuge/km), Szenario IV: 2-streifiger
	 * Messquerschnitt im Verflechtungsbereich.
	 */
	private LinguistischeVariable lvKBSzenario4;

	/**
	 * Bemessungsdichte (KB in Fahrzeuge/km), Szenario V: 3-streifiger
	 * Messquerschnitt im Verflechtungsbereich.
	 */
	private LinguistischeVariable lvKBSzenario5;

	/**
	 * Bemessungsdichte (KB in Fahrzeuge/km), Szenario VI: 4-streifiger
	 * Messquerschnitt im Verflechtungsbereich.
	 */
	private LinguistischeVariable lvKBSzenario6;

	/**
	 * Verkehrsstärke (QKfz) in Anzahl pro Messabschnittsdauer), Szenario I:
	 * 2-streifiger Messquerschnitt.
	 */
	private LinguistischeVariable lvQKfzSzenario1;

	/**
	 * Verkehrsstärke (QKfz) in Anzahl pro Messabschnittsdauer), Szenario I:
	 * 3-streifiger Messquerschnitt.
	 */
	private LinguistischeVariable lvQKfzSzenario2;

	/**
	 * Verkehrsstärke (QKfz) in Anzahl pro Messabschnittsdauer), Szenario I:
	 * 4-streifiger Messquerschnitt.
	 */
	private LinguistischeVariable lvQKfzSzenario3;

	/**
	 * Standardabweichung der Kfz-Geschwindigkeiten (SKfz in km/h).
	 */
	private LinguistischeVariable lvSKfz;

	/**
	 * Geschwindigkeit (VKfz in km/h), Szenario I: 2-streifiger Messquerschnitt
	 * auf freier Strecke.
	 */
	private LinguistischeVariable lvVKfzSzenario1;

	/**
	 * Geschwindigkeit (VKfz in km/h), Szenario II: 3-streifiger Messquerschnitt
	 * auf freier Strecke.
	 */
	private LinguistischeVariable lvVKfzSzenario2;

	/**
	 * Geschwindigkeit (VKfz in km/h), Szenario III: 4-streifiger
	 * Messquerschnitt auf freier Strecke.
	 */
	private LinguistischeVariable lvVKfzSzenario3;

	/**
	 * Geschwindigkeit (VKfz in km/h), Szenario IV: 2-streifiger Messquerschnitt
	 * im Verflechtungsbereich.
	 */
	private LinguistischeVariable lvVKfzSzenario4;

	/**
	 * Geschwindigkeit (VKfz in km/h), Szenario V: 3-streifiger Messquerschnitt
	 * im Verflechtungsbereich.
	 */
	private LinguistischeVariable lvVKfzSzenario5;

	/**
	 * Geschwindigkeit (VKfz in km/h), Szenario VI: 4-streifiger Messquerschnitt
	 * im Verflechtungsbereich.
	 */
	private LinguistischeVariable lvVKfzSzenario6;

	/** Verkehrszustand. */
	private LinguistischeVariable lvVerkehrsZustand;

	/** Unruhe im Verkehrsfluss. */
	private LinguistischeVariable lvVerkehrsZustandUnruhe;

	/** Niederschlagsintensität (Ni in mm/h). */
	private LinguistischeVariable lvNiederschlagsIntensitaet;

	/** Windrichtung (WR in Grad). */
	private LinguistischeVariable lvWindRichtung;

	/** Windgeschwindigkeit (WG in m/s). */
	private LinguistischeVariable lvWindGeschwindigkeitMittelWert;

	/** Sichtweite (SW in m). */
	private LinguistischeVariable lvSichtWeite;

	/** Helligkeit (Helligkeit in Lux). */
	private LinguistischeVariable lvHelligkeit;

	/** Sichtverh&auml;ltnisse. */
	private LinguistischeVariable lvSichtVerhaeltnisse;

	/** Fahrbahnzustand. */
	private LinguistischeVariable lvFahrBahnZustand;

	/** Gl&auml;tte. */
	private LinguistischeVariable lvGlaette;

	/**
	 * Konstruktor verstecken.
	 */
	private FuzzyParameterFactory() {
		// nix
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvALkw} wieder.
	 *
	 * @return {@code lvALkw}.
	 */
	public LinguistischeVariable getLvALkw() {
		if (lvALkw == null) {
			lvALkw = new LinguistischeVariable(ALKW, 0, 100);
			lvALkw.addFuzzySet(new FuzzySet(SEHR_GERING, 0, 0, 3, 7));
			lvALkw.addFuzzySet(new FuzzySet(GERING, 3, 7, 8, 12));
			lvALkw.addFuzzySet(new FuzzySet(MITTEL, 8, 12, 18, 22));
			lvALkw.addFuzzySet(new FuzzySet(HOCH, 18, 22, 23, 27));
			lvALkw.addFuzzySet(new FuzzySet(SEHR_HOCH, 23, 27, 100, 100));
		}
		return lvALkw.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvFahrBahnZustand} wieder.
	 *
	 * @return {@code lvFahrBahnZustand}.
	 */
	public LinguistischeVariable getLvFahrBahnZustand() {
		if (lvFahrBahnZustand == null) {
			lvFahrBahnZustand = new LinguistischeVariable(FAHRBAHNZUSTAND, 0,
					0);
			lvFahrBahnZustand.addFuzzySet(new FuzzySet(TROCKEN, 0, 0, 0, 0));
			lvFahrBahnZustand.addFuzzySet(new FuzzySet(NASS, 0, 0, 0, 0));
			lvFahrBahnZustand.addFuzzySet(new FuzzySet(GLATT, 0, 0, 0, 0));
		}
		return lvFahrBahnZustand.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvGlaette} wieder.
	 *
	 * @return {@code lvGlaette}.
	 */
	public LinguistischeVariable getLvGlaette() {
		if (lvGlaette == null) {
			lvGlaette = new LinguistischeVariable(GLAETTE, 1, 12);
			lvGlaette.addFuzzySet(new FuzzySet(NEIN, 1, 1, 9, 9));
			lvGlaette.addFuzzySet(new FuzzySet(JA, 10, 10, 12, 12));
		}
		return lvGlaette.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvHelligkeit} wieder.
	 *
	 * @return {@code lvHelligkeit}.
	 */
	public LinguistischeVariable getLvHelligkeit() {
		if (lvHelligkeit == null) {
			lvHelligkeit = new LinguistischeVariable(HELLIGKEIT, 0, 60000);
			lvHelligkeit.addFuzzySet(new FuzzySet(DUNKEL, 0, 0, 85, 110));
			lvHelligkeit
					.addFuzzySet(new FuzzySet(DAEMMRIG, 85, 100, 1800, 2100));
			lvHelligkeit
					.addFuzzySet(new FuzzySet(HELL, 1800, 2100, 14500, 15500));
			lvHelligkeit.addFuzzySet(
					new FuzzySet(SEHR_HELL, 14500, 15500, 60000, 60000));
		}
		return lvHelligkeit.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvKBSzenario1} wieder.
	 *
	 * @return {@code lvKBSzenario1}.
	 */
	public LinguistischeVariable getLvKBSzenario1() {
		if (lvKBSzenario1 == null) {
			lvKBSzenario1 = new LinguistischeVariable(KB, 0, 10000);
			lvKBSzenario1.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 25, 35));
			lvKBSzenario1
					.addFuzzySet(new FuzzySet(NICHT_SEHR_HOCH, 0, 0, 50, 70));
			lvKBSzenario1.addFuzzySet(new FuzzySet(HOCH, 25, 35, 50, 70));
			lvKBSzenario1
					.addFuzzySet(new FuzzySet(SEHR_HOCH, 50, 70, 10000, 10000));
		}
		return lvKBSzenario1.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvKBSzenario2} wieder.
	 *
	 * @return {@code lvKBSzenario2}.
	 */
	public LinguistischeVariable getLvKBSzenario2() {
		if (lvKBSzenario2 == null) {
			lvKBSzenario2 = new LinguistischeVariable(KB, 0, 10000);
			lvKBSzenario2.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 35, 45));
			lvKBSzenario2
					.addFuzzySet(new FuzzySet(NICHT_SEHR_HOCH, 0, 0, 60, 80));
			lvKBSzenario2.addFuzzySet(new FuzzySet(HOCH, 35, 45, 60, 80));
			lvKBSzenario2
					.addFuzzySet(new FuzzySet(SEHR_HOCH, 60, 80, 10000, 10000));
		}
		return lvKBSzenario2.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvKBSzenario3} wieder.
	 *
	 * @return {@code lvKBSzenario3}.
	 */
	public LinguistischeVariable getLvKBSzenario3() {
		if (lvKBSzenario3 == null) {
			lvKBSzenario3 = new LinguistischeVariable(KB, 0, 10000);
			lvKBSzenario3.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 35, 45));
			lvKBSzenario3
					.addFuzzySet(new FuzzySet(NICHT_SEHR_HOCH, 0, 0, 45, 55));
			lvKBSzenario3.addFuzzySet(new FuzzySet(HOCH, 35, 45, 60, 80));
			lvKBSzenario3
					.addFuzzySet(new FuzzySet(SEHR_HOCH, 45, 55, 10000, 10000));
		}
		return lvKBSzenario3.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvKBSzenario4} wieder.
	 *
	 * @return {@code lvKBSzenario4}.
	 */
	public LinguistischeVariable getLvKBSzenario4() {
		if (lvKBSzenario4 == null) {
			lvKBSzenario4 = new LinguistischeVariable(KB, 0, 10000);
			lvKBSzenario4.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 15, 25));
			lvKBSzenario4
					.addFuzzySet(new FuzzySet(NICHT_SEHR_HOCH, 0, 0, 40, 60));
			lvKBSzenario4.addFuzzySet(new FuzzySet(HOCH, 15, 25, 40, 60));
			lvKBSzenario4
					.addFuzzySet(new FuzzySet(SEHR_HOCH, 40, 60, 10000, 10000));
		}
		return lvKBSzenario4.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvKBSzenario5} wieder.
	 *
	 * @return {@code lvKBSzenario5}.
	 */
	public LinguistischeVariable getLvKBSzenario5() {
		if (lvKBSzenario5 == null) {
			lvKBSzenario5 = new LinguistischeVariable(KB, 0, 10000);
			lvKBSzenario5.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 25, 35));
			lvKBSzenario5
					.addFuzzySet(new FuzzySet(NICHT_SEHR_HOCH, 0, 0, 50, 70));
			lvKBSzenario5.addFuzzySet(new FuzzySet(HOCH, 25, 35, 50, 70));
			lvKBSzenario5
					.addFuzzySet(new FuzzySet(SEHR_HOCH, 50, 70, 10000, 10000));
		}
		return lvKBSzenario5.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvKBSzenario6} wieder.
	 *
	 * @return {@code lvKBSzenario6}.
	 */
	public LinguistischeVariable getLvKBSzenario6() {
		if (lvKBSzenario6 == null) {
			lvKBSzenario6 = new LinguistischeVariable(KB, 0, 10000);
			lvKBSzenario6.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 35, 45));
			lvKBSzenario6
					.addFuzzySet(new FuzzySet(NICHT_SEHR_HOCH, 0, 0, 60, 80));
			lvKBSzenario6.addFuzzySet(new FuzzySet(HOCH, 35, 45, 60, 80));
			lvKBSzenario6
					.addFuzzySet(new FuzzySet(SEHR_HOCH, 60, 80, 10000, 10000));
		}
		return lvKBSzenario6.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvNiederschlagsIntensitaet} wieder.
	 *
	 * @return {@code lvNiederschlagsIntensitaet}.
	 */
	public LinguistischeVariable getLvNiederschlagsIntensitaet() {
		if (lvNiederschlagsIntensitaet == null) {
			lvNiederschlagsIntensitaet = new LinguistischeVariable(NI, 0, 200);
			lvNiederschlagsIntensitaet
					.addFuzzySet(new FuzzySet(KEIN_NI, 0, 0, 0, 1));
			lvNiederschlagsIntensitaet
					.addFuzzySet(new FuzzySet(SCHWACHER_NI, 0, 1, 4, 6));
			lvNiederschlagsIntensitaet
					.addFuzzySet(new FuzzySet(MITTLERER_NI, 4, 6, 8, 12));
			lvNiederschlagsIntensitaet
					.addFuzzySet(new FuzzySet(STARKER_NI, 8, 12, 14, 16));
			lvNiederschlagsIntensitaet.addFuzzySet(
					new FuzzySet(SEHR_STARKER_NI, 14, 16, 200, 200));
		}
		return lvNiederschlagsIntensitaet.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvQKfzSzenario1} wieder.
	 *
	 * @return {@code lvQKfzSzenario1}.
	 */
	public LinguistischeVariable getLvQKfzSzenario1() {
		if (lvQKfzSzenario1 == null) {
			lvQKfzSzenario1 = new LinguistischeVariable(QKFZ, 0, 100000000);
			lvQKfzSzenario1
					.addFuzzySet(new FuzzySet(SEHR_NIEDRIG, 0, 0, 500, 700));
			lvQKfzSzenario1
					.addFuzzySet(new FuzzySet(NIEDRIG, 500, 700, 1000, 1400));
			lvQKfzSzenario1
					.addFuzzySet(new FuzzySet(MITTEL, 1000, 1400, 2000, 2400));
			lvQKfzSzenario1
					.addFuzzySet(new FuzzySet(HOCH, 2000, 2400, 3000, 3400));
			lvQKfzSzenario1.addFuzzySet(
					new FuzzySet(SEHR_HOCH, 3000, 3400, 100000000, 100000000));
		}
		return lvQKfzSzenario1.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvQKfzSzenario2} wieder.
	 *
	 * @return {@code lvQKfzSzenario2}.
	 */
	public LinguistischeVariable getLvQKfzSzenario2() {
		if (lvQKfzSzenario2 == null) {
			lvQKfzSzenario2 = new LinguistischeVariable(QKFZ, 0, 100000000);
			lvQKfzSzenario2
					.addFuzzySet(new FuzzySet(SEHR_NIEDRIG, 0, 0, 750, 1050));
			lvQKfzSzenario2
					.addFuzzySet(new FuzzySet(NIEDRIG, 750, 1050, 1500, 2100));
			lvQKfzSzenario2
					.addFuzzySet(new FuzzySet(MITTEL, 1500, 2100, 3000, 36000));
			lvQKfzSzenario2
					.addFuzzySet(new FuzzySet(HOCH, 3000, 3600, 4500, 5100));
			lvQKfzSzenario2.addFuzzySet(
					new FuzzySet(SEHR_HOCH, 4500, 5100, 100000000, 100000000));
		}
		return lvQKfzSzenario2.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvQKfzSzenario3} wieder.
	 *
	 * @return {@code lvQKfzSzenario3}.
	 */
	public LinguistischeVariable getLvQKfzSzenario3() {
		if (lvQKfzSzenario3 == null) {
			lvQKfzSzenario3 = new LinguistischeVariable(QKFZ, 0, 100000000);
			lvQKfzSzenario3
					.addFuzzySet(new FuzzySet(SEHR_NIEDRIG, 0, 0, 1000, 1400));
			lvQKfzSzenario3
					.addFuzzySet(new FuzzySet(NIEDRIG, 1000, 1400, 2000, 2800));
			lvQKfzSzenario3
					.addFuzzySet(new FuzzySet(MITTEL, 2000, 2800, 4000, 48000));
			lvQKfzSzenario3
					.addFuzzySet(new FuzzySet(HOCH, 4000, 4800, 6000, 6800));
			lvQKfzSzenario3.addFuzzySet(
					new FuzzySet(SEHR_HOCH, 6000, 6800, 100000000, 100000000));
		}
		return lvQKfzSzenario3.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvSichtVerhaeltnisse} wieder.
	 *
	 * @return {@code lvSichtVerhaeltnisse}.
	 */
	public LinguistischeVariable getLvSichtVerhaeltnisse() {
		if (lvSichtVerhaeltnisse == null) {
			lvSichtVerhaeltnisse = new LinguistischeVariable(SICHT, 0, 0);
			lvSichtVerhaeltnisse.addFuzzySet(new FuzzySet(GUT, 0, 0, 0, 0));
			lvSichtVerhaeltnisse.addFuzzySet(
					new FuzzySet(LEICHT_EINGESCHRAENKT, 0, 0, 0, 0));
			lvSichtVerhaeltnisse
					.addFuzzySet(new FuzzySet(EINGESCHRAENKT, 0, 0, 0, 0));
			lvSichtVerhaeltnisse
					.addFuzzySet(new FuzzySet(UNZUREICHEND, 0, 0, 0, 0));
		}
		return lvSichtVerhaeltnisse.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvSichtWeite} wieder.
	 *
	 * @return {@code lvSichtWeite}.
	 */
	public LinguistischeVariable getLvSichtWeite() {
		if (lvSichtWeite == null) {
			lvSichtWeite = new LinguistischeVariable(SICHTWEITE, 10, 2000);
			lvSichtWeite
					.addFuzzySet(new FuzzySet(SICHT_BIS_50M, 10, 10, 45, 55));
			lvSichtWeite
					.addFuzzySet(new FuzzySet(SICHT_BIS_80M, 45, 55, 75, 85));
			lvSichtWeite.addFuzzySet(
					new FuzzySet(SICHT_BIS_120M, 75, 85, 110, 130));
			lvSichtWeite.addFuzzySet(
					new FuzzySet(SICHT_BIS_250M, 110, 130, 230, 270));
			lvSichtWeite.addFuzzySet(
					new FuzzySet(SICHT_BIS_400M, 230, 270, 370, 430));
			lvSichtWeite.addFuzzySet(
					new FuzzySet(SICHT_UEBER_400M, 370, 430, 2000, 2000));
		}
		return lvSichtWeite.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvSKfz} wieder.
	 *
	 * @return {@code lvSKfz}.
	 */
	public LinguistischeVariable getLvSKfz() {
		if (lvSKfz == null) {
			lvSKfz = new LinguistischeVariable(SKFZ, 0, 254);
			lvSKfz.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 18, 22));
			lvSKfz.addFuzzySet(new FuzzySet(HOCH, 18, 22, 254, 254));
		}
		return lvSKfz.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVerkehrsZustand} wieder.
	 *
	 * @return {@code lvVerkehrsZustand}.
	 */
	public LinguistischeVariable getLvVerkehrsZustand() {
		if (lvVerkehrsZustand == null) {
			lvVerkehrsZustand = new LinguistischeVariable(VERKEHRSZUSTAND, 0,
					0);
			lvVerkehrsZustand.addFuzzySet(new FuzzySet(GESTAUT, 0, 0, 0, 0));
			lvVerkehrsZustand.addFuzzySet(new FuzzySet(ZAEH, 0, 0, 0, 0));
			lvVerkehrsZustand.addFuzzySet(new FuzzySet(DICHT, 0, 0, 0, 0));
			lvVerkehrsZustand.addFuzzySet(new FuzzySet(FREI, 0, 0, 0, 0));
		}
		return lvVerkehrsZustand.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVerkehrsZustandUnruhe} wieder.
	 *
	 * @return {@code lvVerkehrsZustandUnruhe}.
	 */
	public LinguistischeVariable getLvVerkehrsZustandUnruhe() {
		if (lvVerkehrsZustandUnruhe == null) {
			lvVerkehrsZustandUnruhe = new LinguistischeVariable(
					VERKEHRSZUSTAND_UNRUHE, 0, 0);
			lvVerkehrsZustandUnruhe
					.addFuzzySet(new FuzzySet(UNRUHE, 0, 0, 0, 0));
			lvVerkehrsZustandUnruhe
					.addFuzzySet(new FuzzySet(KEINE_UNRUHE, 0, 0, 0, 0));
		}
		return lvVerkehrsZustandUnruhe.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVKfzSzenario1} wieder.
	 *
	 * @return {@code lvVKfzSzenario1}.
	 */
	public LinguistischeVariable getLvVKfzSzenario1() {
		if (lvVKfzSzenario1 == null) {
			lvVKfzSzenario1 = new LinguistischeVariable(VKFZ, 0, 254);
			lvVKfzSzenario1.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 25, 35));
			lvVKfzSzenario1.addFuzzySet(new FuzzySet(MITTEL, 25, 35, 75, 85));
			lvVKfzSzenario1.addFuzzySet(new FuzzySet(HOCH, 75, 85, 254, 254));
		}
		return lvVKfzSzenario1.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVKfzSzenario2} wieder.
	 *
	 * @return {@code lvVKfzSzenario2}.
	 */
	public LinguistischeVariable getLvVKfzSzenario2() {
		if (lvVKfzSzenario2 == null) {
			lvVKfzSzenario2 = new LinguistischeVariable(VKFZ, 0, 254);
			lvVKfzSzenario2.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 27, 36));
			lvVKfzSzenario2.addFuzzySet(new FuzzySet(MITTEL, 27, 36, 77, 86));
			lvVKfzSzenario2.addFuzzySet(new FuzzySet(HOCH, 77, 86, 254, 254));
		}
		return lvVKfzSzenario2.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVKfzSzenario3} wieder.
	 *
	 * @return {@code lvVKfzSzenario3}.
	 */
	public LinguistischeVariable getLvVKfzSzenario3() {
		if (lvVKfzSzenario3 == null) {
			lvVKfzSzenario3 = new LinguistischeVariable(VKFZ, 0, 254);
			lvVKfzSzenario3.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 28, 36));
			lvVKfzSzenario3.addFuzzySet(new FuzzySet(MITTEL, 28, 36, 78, 86));
			lvVKfzSzenario3.addFuzzySet(new FuzzySet(HOCH, 78, 86, 254, 254));
		}
		return lvVKfzSzenario3.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVKfzSzenario4} wieder.
	 *
	 * @return {@code lvVKfzSzenario4}.
	 */
	public LinguistischeVariable getLvVKfzSzenario4() {
		if (lvVKfzSzenario4 == null) {
			lvVKfzSzenario4 = new LinguistischeVariable(VKFZ, 0, 254);
			lvVKfzSzenario4.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 15, 30));
			lvVKfzSzenario4.addFuzzySet(new FuzzySet(MITTEL, 15, 30, 60, 80));
			lvVKfzSzenario4.addFuzzySet(new FuzzySet(HOCH, 60, 80, 254, 254));
		}
		return lvVKfzSzenario4.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVKfzSzenario5} wieder.
	 *
	 * @return {@code lvVKfzSzenario5}.
	 */
	public LinguistischeVariable getLvVKfzSzenario5() {
		if (lvVKfzSzenario5 == null) {
			lvVKfzSzenario5 = new LinguistischeVariable(VKFZ, 0, 254);
			lvVKfzSzenario5.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 20, 30));
			lvVKfzSzenario5.addFuzzySet(new FuzzySet(MITTEL, 20, 30, 70, 80));
			lvVKfzSzenario5.addFuzzySet(new FuzzySet(HOCH, 70, 80, 254, 254));
		}
		return lvVKfzSzenario5.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvVKfzSzenario6} wieder.
	 *
	 * @return {@code lvVKfzSzenario6}.
	 */
	public LinguistischeVariable getLvVKfzSzenario6() {
		if (lvVKfzSzenario6 == null) {
			lvVKfzSzenario6 = new LinguistischeVariable(VKFZ, 0, 254);
			lvVKfzSzenario6.addFuzzySet(new FuzzySet(NIEDRIG, 0, 0, 23, 33));
			lvVKfzSzenario6.addFuzzySet(new FuzzySet(MITTEL, 23, 33, 73, 83));
			lvVKfzSzenario6.addFuzzySet(new FuzzySet(HOCH, 73, 83, 254, 254));
		}
		return lvVKfzSzenario6.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvWindGeschwindigkeitMittelWert}
	 * wieder.
	 *
	 * @return {@code lvWindGeschwindigkeitMittelWert}.
	 */
	public LinguistischeVariable getLvWindGeschwindigkeitMittelWert() {
		if (lvWindGeschwindigkeitMittelWert == null) {
			lvWindGeschwindigkeitMittelWert = new LinguistischeVariable(WG, 0,
					100);
			lvWindGeschwindigkeitMittelWert
					.addFuzzySet(new FuzzySet(KEIN_WIND, 0, 0, 0, 1));
			lvWindGeschwindigkeitMittelWert
					.addFuzzySet(new FuzzySet(SCHWACHER_WIND, 0, 1, 3, 4));
			lvWindGeschwindigkeitMittelWert
					.addFuzzySet(new FuzzySet(MITTLERER_WIND, 3, 4, 8, 10));
			lvWindGeschwindigkeitMittelWert
					.addFuzzySet(new FuzzySet(STARKER_WIND, 8, 10, 100, 100));
		}
		return lvWindGeschwindigkeitMittelWert.clone();
	}

	/**
	 * Gibt den Wert der Eigenschaft {@code lvWindRichtung} wieder.
	 *
	 * @return {@code lvWindRichtung}.
	 */
	public LinguistischeVariable getLvWindRichtung() {
		if (lvWindRichtung == null) {
			lvWindRichtung = new LinguistischeVariable(WR, 0, 359);
			lvWindRichtung.addFuzzySet(new FuzzySet(N, 335, 340, 20, 25));
			lvWindRichtung.addFuzzySet(new FuzzySet(NO, 20, 25, 65, 70));
			lvWindRichtung.addFuzzySet(new FuzzySet(O, 65, 70, 110, 115));
			lvWindRichtung.addFuzzySet(new FuzzySet(SO, 110, 115, 155, 160));
			lvWindRichtung.addFuzzySet(new FuzzySet(S, 155, 160, 200, 205));
			lvWindRichtung.addFuzzySet(new FuzzySet(SW, 200, 205, 290, 295));
			lvWindRichtung.addFuzzySet(new FuzzySet(W, 245, 250, 290, 295));
			lvWindRichtung.addFuzzySet(new FuzzySet(NW, 290, 295, 335, 340));
		}
		return lvWindRichtung.clone();
	}

}
